// (C) DownThemAll! - See COPYING/SOURCE

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const log = Components.utils.reportError;
const TOPIC_FILTERSCHANGED = 'DTA:filterschanged';
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

String.prototype.count = function(s) {
return Array.reduce(this, function(c, a) c + (s.indexOf(a) != -1 ? 1 : 0), 0);
}
function commonprefix(m) {
let [s1, s2] = m;
let n = Math.min(s1.length, s2.length);
for (let i = 0; i < n; ++i) {
if (s1[i] != s2[i]) {
return s1.slice(0, i);
}
}
return s1.slice(0, n);
}
function icombine(arr) {
for (let i = 0, e = arr.length; i < e - 1; ++i) {
for (let j = i + 1; j < e; ++j) {
yield [arr[i], arr[j]];
}
}
}
function combine(arr) {
return [a for (a in icombine(arr))];
}
function biggestgroup(slist, fngroup) {
let d = {};
for each (let x in combine(slist)) {
let k = fngroup(x);
if (!k || k.count('(') != k.count(')')) {
continue;
}
if (!(k in d)) {
d[k] = x;
}
else {
k = d[k];
x.forEach(function(e) k.push(e));
}
}
let fix = null;
let max = 0;
for (let k in d) {
let m = k.length;
if (max < m) {
max = m;
fix = k;
}
}
if (!fix) {
return [null, null];
}
let rlist = [];
for each (let i in d[fix]) {
if (rlist.indexOf(i) == -1) {
rlist.push(i);
}
}
return [fix, rlist];
}
function merge(slist) {
while (slist.length) {
let [pre, prelist] = biggestgroup(slist, commonprefix);
if (!pre) {
break;
}
slist = slist.filter(function(e) prelist.indexOf(e) == -1);
prelist = prelist.map(function(e) e.slice(pre.length));
if (prelist.length > 1) {
slist.push(pre + "(?:" + prelist.join("|") + ")");
}
else {
slist.push(pre + prelist.join(""));
}
}
return slist.join("|");
}

function AutoFilter() {};
AutoFilter.prototype = {
classDescription: "DownThemAll! AutoContainer automated filter creator",
classID: Components.ID('a650c130-22ec-11de-8c30-0800200c9a66'),
contractID: '@downthemall.net/anticontainer/autofilter;1',
_xpcom_categories: [{category: 'app-startup'}],

QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference, Ci.nsIWeakReference]),
QueryReferent: function(iid) this.QueryInterface(iid),
GetWeakReference: function() this,
get _os() {
return Cc['@mozilla.org/observer-service;1']
.getService(Ci.nsIObserverService);
},
get _fm() {
return Cc['@downthemall.net/filtermanager;2']
.getService(Components.interfaces.dtaIFilterManager);
},
get plugins() {
let plgs = {};
Components.utils.import('resource://dtaac/plugins.jsm', plgs);
delete AutoFilter.prototype.plugins;
return this.plugins = AutoFilter.prototype.plugins = plgs;
},
get allPlugins() {
return [p.strmatch for (p in this.plugins.enumerate())];
},
init: function af_init() {
this._os.addObserver(this, 'xpcom-shutdown', false);
this._os.addObserver(this, 'final-ui-startup', false);
this._os.addObserver(this, this.plugins.TOPIC_PLUGINSCHANGED, false);
this._os.addObserver(this, TOPIC_FILTERSCHANGED, false);
},
dispose: function af_dispose() {
this._os.removeObserver(this, 'xpcom-shutdown');
this._os.removeObserver(this, this.plugins.TOPIC_PLUGINSCHANGED);
this._os.removeObserver(this, TOPIC_FILTERSCHANGED);
},
reload: function af_reload() {
try {
let merged = '/' + this.allPlugins.map(function(r) '(?:' + r + ')').join('|').replace(/\//g, '\\/') + '/i';


let f;
try {
f = this._fm.getFilter('deffilter-ac');
}
catch (ex) {
log("dtaac: autofilter reload < 1.1.3 compat");
try {
f = this._fm.getFilter('extensions.dta.filters.deffilter-ac');
}
catch (ex) {
log(ex);
return;
}
}
if (f.expression != merged) {
f.expression = merged;
f.save();
}
}
catch (ex) {
log(ex);
}
},
observe: function af_observe(subject, topic, data) {
switch (topic) {
case 'xpcom-shutdown':
this.dispose();
break;
case 'app-startup':
try {
this._os.removeObserver(this, 'app-startup');
}
catch (ex) {  }

this.init();
break;
case 'final-ui-startup':
this._os.removeObserver(this, 'final-ui-startup');

this._fm.reload();
break;
case this.plugins.TOPIC_PLUGINSCHANGED:
case TOPIC_FILTERSCHANGED:
this.reload();
break;
}
}
};

function NSGetModule(compMgr, fileSpec) {
return XPCOMUtils.generateModule([AutoFilter]);
}