// (C) DownThemAll! - See COPYING/SOURCE
const EXPORTED_SYMBOLS = [
'pushPlugin', 'popPlugin',
'nsJSON',
'loadPluginFromStream', 'loadPluginFromFile',
'enumerate',
'installFromFile', 'installFromWeb', 'uninstallPlugin', 'createNewPlugin',
'prettyJSON',
'TOPIC_PLUGINSCHANGED', 'DEFAULT_NAMESPACE'
];
const Cc = Components.classes;
const Ci = Components.interfaces;
const log = Components.utils.reportError;
const TOPIC_PLUGINSCHANGED = 'DTA:AC:pluginschanged';
const DEFAULT_NAMESPACE = 'nonymous';
const ConverterOutputStream = Components.Constructor('@mozilla.org/intl/converter-output-stream;1', 'nsIConverterOutputStream', 'init');
const FileInputStream = Components.Constructor('@mozilla.org/network/file-input-stream;1', 'nsIFileInputStream', 'init');
const FileOutputStream = Components.Constructor('@mozilla.org/network/file-output-stream;1', 'nsIFileOutputStream', 'init');
const File = new Components.Constructor('@mozilla.org/file/local;1', 'nsILocalFile', 'initWithPath');
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

this.__defineGetter__('Prefs', function() {
let p = Cc['@mozilla.org/preferences-service;1']
.getService(Ci.nsIPrefService)
.getBranch('extensions.dta.')
.QueryInterface(Ci.nsIPrefBranch2);
delete this.Prefs;
return this.Prefs = p;
});
this.__defineGetter__('EM_DIR', function() {
let ed = Cc['@mozilla.org/extensions/manager;1']
.getService(Ci.nsIExtensionManager)
.getInstallLocation('anticontainer@downthemall.net')
.getItemFile('anticontainer@downthemall.net', 'plugins/');
delete this.EM_DIR;
return this.EM_DIR = ed;
});
this.__defineGetter__('PD_DIR', function() {
let pd = Cc['@mozilla.org/file/directory_service;1']
.getService(Ci.nsIProperties)
.get("ProfD", Ci.nsILocalFile);
delete this.PD_DIR;
return this.PD_DIR = pd;
});

this.__defineGetter__('USER_DIR', function() {
let d = PD_DIR.clone();
d.append('anticontainer_plugins');
if (!d.exists()) {
d.create(Ci.nsIFile.DIRECTORY_TYPE, 0664);
}
delete this.USER_DIR;
return this.USER_DIR = d;
});
this.__defineGetter__('nsJSON', function() {
delete this.nsJSON;
return this.nsJSON = Cc['@mozilla.org/dom/json;1'].createInstance(Ci.nsIJSON);
});
this.__defineGetter__('UUID', function() {
let ug = Cc["@mozilla.org/uuid-generator;1"].getService(Ci.nsIUUIDGenerator);
delete this.UUID;
return this.UUID = ug;
});
function newUUID() UUID.generateUUID().toString();
function Observer() {
Prefs.addObserver('anticontainer.disabled_plugins', this, true);
Prefs.addObserver('filters.deffilter-ac', this, true);
}
Observer.prototype = {
_os: Cc['@mozilla.org/observer-service;1'].getService(Ci.nsIObserverService),
QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference, Ci.nsIWeakReference]),
QueryReferent: function(iid) this.QueryInterface(iid),
GetWeakReference: function() this,
observe: function() {
this.notify();
},
notify: function() {
this._os.notifyObservers(null, TOPIC_PLUGINSCHANGED, null);
}
};
const observer = new Observer();
let lastFilters = 0;
function validatePlugin(o) {
if (['redirector', 'resolver', 'sandbox'].indexOf(o.type) == -1) {
throw new Error("Failed to load plugin: invalid type");
}
switch (o.type) {
case 'resolver':
if (!o.finder || !o.builder) {
throw new Error("Failed to load plugin: incomplete resolver!");
}
break;
case 'redirector':
if (!o.pattern || !o.match) {
throw new Error("Failed to load plugin: incomplete redirector!");
}
break;
case 'sandbox':
if (!o.process && !o.resolve) {
throw new Error("Failed to load plugin: sandboxed plugin doesn't implement anything!");
}
break;
}
if (!o.prefix || typeof o.prefix != 'string') {
throw new Error("Failed to load plugin: prefix omitted");
}
o.source = nsJSON.encode(o);
for each (let x in ['match', 'finder', 'pattern']) {
if (x in o) {
o['str' + x] = o[x];
o[x] = new RegExp(o[x], 'im');
}
}
for each (let c in o.cleaners) {
for each (let x in ['pattern']) {
if (x in c) {
c['str' + x] = c[x];
c[x] = new RegExp(c[x], 'i');
}
}
}
for each (let b in ['static', 'decode', 'static', 'omitReferrer', 'sendInitialReferrer', 'useServerName']) {
o[b] = !!o[b];
}
if (!o.priority || typeof o.priority != 'number') {
o.priority = 0;
}
o.priority = Math.round(o.priority);
if (!o.ns) {
o.ns = DEFAULT_NAMESPACE;
}
o.ns = o.ns.toString();
o.id = o.prefix + '@' + o.ns;
return o;
}

function loadPluginFromStream(stream, size) {
return validatePlugin(nsJSON.decodeFromStream(stream, size));
}

function loadPluginFromFile(file) {
if (!(file instanceof Ci.nsIFile)) {
file = new File(file);
}
let fs = new FileInputStream(file, 0x01, 0, 1<<2);
let o = loadPluginFromStream(fs, file.size);
fs.close();
o.file = file;
return o;
}
function idToFilename(id) id.replace(/[^\w\d\._@-]/gi, '-') + ".json";
function _enumerate(enumerators, p) {
if (!(p instanceof Array)) {
p = [];
}
let i = 0;
for each (let [managed, prio, e] in enumerators) {
while (e.hasMoreElements()) {
let f = e.getNext().QueryInterface(Ci.nsIFile);
if (f.leafName.search(/\.json$/i) != -1) {
try {
let o = loadPluginFromFile(f);
if (p.indexOf(o.id) != -1) {
continue;
}
o.priority += prio;
o.managed = managed;
++i;
yield o;
}
catch (ex) {
log("Failed to load " + f.leafName);
log(ex);
}
}
}
}
if (lastFilters && i != lastFilters) {
log('dtaac:plugins: notify because of new numPlugins');
lastFilters = i;
observer.notify();
}
}

function enumerate(all) {
let enums = [[true, 1, EM_DIR.directoryEntries]];
try {
enums.push([false, 3, USER_DIR.directoryEntries]);
}
catch (ex) {
}
let g = _enumerate(enums, all ? [] : uneval(Prefs.getCharPref('anticontainer.disabled_plugins')));
for (let e in g) {
yield e;
}
}

function installFromFile(file) {
let p = loadPluginFromFile(file);
let pd = USER_DIR.clone();
let nn = idToFilename(p.id);
file.copyTo(pd, nn);
pd.append(nn);
p.file = pd;
observer.notify();
return p;
}
function installFromStringOrObject(str) {
str = prettyJSON(str);
let p = validatePlugin(nsJSON.decode(str));
let pf = USER_DIR.clone();
pf.append(idToFilename(p.id));
let cs = ConverterOutputStream(
new FileOutputStream(pf, 0x02 | 0x08 | 0x20, -1, 0),
null,
0,
null
);
cs.writeString(str);
cs.close();
observer.notify();
return {id: p.id, file: pf};
}

function installFromWeb(str, updateURL) {
let p = nsJSON.decode(str);
p.fromWeb = true;
if (!p.updateURL && updateURL) {
p.updateURL = updateURL;
}
return installFromStringOrObject(p);
}

function uninstallPlugin(id) {
let pf = USER_DIR.clone();
pf.append(idToFilename(id));
if (!pf.exists()) {
throw new Error("Cannot find plugin for id: " + id + ", tried: " + pf.path);
}
pf.remove(false);
observer.notify();
}
function createNewPlugin(plugin) {
switch (plugin.type) {
case 'redirector':
plugin.pattern = "<fill>";
plugin.replacement = "<fill>";
break;
case 'resolver':
plugin.finder = "<fill regexp>";
plugin.builder = "<fill builder>";
break;
case 'sandbox':
process.process = 'makeRequest(baseURL, resolve, resolve);';
process.resolve = 'defaultResolve();';
break;
}
return installFromStringOrObject(plugin);
}
const _store = {};

function pushPlugin(id, plug) {
_store[id] = plug;
}

function popPlugin(id) {
if (!(id in _store)) {
throw new Error("plugin not found!");
}
let rv = _store[id];
delete _store[id];
return rv;
}

function prettyJSON(objectOrString, initialIndent) {

function prettyPrint(_oo, _l) {
let _p = [];
function l(n) {
let rv = '';
while(--n >= 0) {
rv += '\t';
}
return rv;
}
function p(o, _l) {
if (typeof o == "string") {
o = o.replace(/[\\"\x00-\x1F\u0080-\uFFFF]/g, function($0) {
switch ($0) {case "\b": return "\\b"; case "\t": return "\\t"; case "\n": return "\\n"; case "\f": return "\\f"; case "\r": return "\\r"; case '"':	return '\\"'; case "\\": return "\\\\";}
return "\\u" + ("0000" + $0.charCodeAt(0).toString(16)).slice(-4);
});
_p.push('"' + o + '"')
}
else if (typeof o == "boolean") {
_p.push(o ? "true" : "false");
}
else if (typeof o == "number" && isFinite(o)) {
_p.push(o.toString());
}
else if (o === null) {
_p.push("null");
}
else if (
o instanceof Array ||
typeof o == "object" && "length" in o &&
(o.length === 0 || o[o.length - 1] !== undefined)
) {
_p.push("[\n");
for (var i = 0; i < o.length; i++) {
arguments.callee(o[i], _l + 1);
_p.push(",");
}
if (o.length > 0)
_p.pop();
_p.push("\n" + l(_l) + "]");
}
else if (typeof o == "object") {
_p.push(l(_l));
_p.push("{\n");
for (var key in o) {
_p.push(l(_l + 1));
arguments.callee(key.toString());
_p.push(": ");
arguments.callee(o[key], _l + 1);
_p.push(",\n");
}
if (_p[_p.length - 1] == ",\n")
_p.pop();
_p.push("\n" + l(_l) + "}");
}
else {
throw new TypeError("No JSON representation for this object!");
}
}
p(_oo, _l ? _l : 0);
return _p.join("");
}
if (typeof objectOrString != 'string') {
objectOrString = nsJSON.encode(objectOrString);
}
return prettyPrint(nsJSON.decode(objectOrString), initialIndent);
}